/* Client for IHello
 * Heavilly modified from:
 */
/*
 * SELFREG.CPP
 * Server Self-Registrtation Utility, Chapter 5
 *
 * Copyright (c)1993-1995 Microsoft Corporation, All Rights Reserved
 *
 * Kraig Brockschmidt, Microsoft
 * Internet  :  kraigb@microsoft.com
 * Compuserve:  >INTERNET:kraigb@microsoft.com
 */


import std.c.stdio;
import std.c.stdlib;
import std.c.windows.windows;
import std.c.windows.com;

import object;
import std.asserterror;

extern(Windows) BOOL FreeLibrary(
  HMODULE hLibModule   // handle to loaded library module
);

extern (Windows) alias HRESULT (*registerFuncPtr)();


class SelfRegisteringDLL {
	static const char * REG_FUNC_NAME   = "DllRegisterServer";
	static const char * UNREG_FUNC_NAME = "DllUnregisterServer";
	HINSTANCE       hmodule;
	registerFuncPtr dllRegister;
	registerFuncPtr dllUnregister;

	this( char[] path ) {
		hmodule = LoadLibraryA(path);
		printf("dll module = %d\n", hmodule );
		if ( hmodule > cast(HINSTANCE)HINSTANCE_ERROR )
		{
			printf("LoadLibraryA() succeeded\n");
			dllRegister   = GetProcAddress( hmodule, REG_FUNC_NAME );
			dllUnregister = GetProcAddress( hmodule, UNREG_FUNC_NAME );
			printf("register   = %p, fn = '%s'\n", dllRegister, REG_FUNC_NAME);
			printf("unregister = %p, fn = '%s'\n", dllUnregister, UNREG_FUNC_NAME);
			if ( (!dllRegister) || (!dllUnregister) ) {
				throw new Exception( "Unable to get required function addresses" );
			}
			if ( !SUCCEEDED( (*dllRegister)() ) ) {
				throw new Exception( "Unable to register" );
			}
		}
		else
		{
			throw new Exception( "Unable to get required dll" );
		}
	}

	~this() {
		if ( !SUCCEEDED( (*dllUnregister)() ) ) {
			throw new Exception( "Unable to unregister" );
		}
		FreeLibrary( hmodule );
		hmodule = null;
	}
}

class CoHelper {
	DWORD   dwVer;
	this() {
		// Make sure COM is the right version
		dwVer = CoBuildVersion();
		if ( rmm != HIWORD(dwVer) )
		{
			throw new Exception( "Incorrect OLE 2 version number" );
		}
		//Initialize OLE
		if ( FAILED( CoInitialize( null ) ) )
		{
			throw new Exception( "OLE 2 failed to initialize" );
		}
		printf("OLE 2 initialized\n");
	}
	~this() {
		CoFreeUnusedLibraries();
		CoUninitialize();
	}
}

import ihello;


template ComCreate(alias clz_guid, alias if_guid, IFACE ) {
	IFACE newInstance() {
		HRESULT hr;
		IFACE   obj;
		hr = CoCreateInstance( &clz_guid, null, CLSCTX_ALL, &if_guid, &obj );
		if ( FAILED( hr ) )
		{
			throw new Exception( "unable to create new instance" );
		}
		return obj;
	}
}
template ComCreate( IFACE ) {
	IFACE newInstance(GUID * clz_guid, GUID * if_guid ) {
		HRESULT hr;
		IFACE   obj;
		hr = CoCreateInstance( clz_guid, null, CLSCTX_ALL, if_guid, &obj );
		if ( FAILED( hr ) )
		{
			throw new Exception( "unable to create new instance" );
		}
		return obj;
	}
}

int main()
{
	IHello	pIHello;
	HRESULT hr;
	try {
		auto CoHelper cohelp = new CoHelper();
		auto SelfRegisteringDLL dll = new SelfRegisteringDLL( "dserver.dll" );
		printf("server registered\n");
		pIHello = ComCreate!(IHello).newInstance(&CLSID_Hello, &IID_IHello);
//		pIHello = ComCreate!(CLSID_Hello, IID_IHello, IHello).newInstance();
//
// enabling the above line causes the following error
//
// dclient.obj(dclient)
//  Error 42: Symbol Undefined _D7dclient10IID_IHelloS3std1c7windows3com4GUID
// dclient.obj(dclient)
//  Error 42: Symbol Undefined _D7dclient11CLSID_HelloS3std1c7windows3com4GUID
//
		printf("Object created, calling IHello.Print(), IHello = %p\n", pIHello);
		pIHello.Print();
		printf("Printf called\n");
		int i = pIHello.Release();
		printf("Release called (%d)\n", i );
	} catch ( Exception e ) {
		printf( "Caught Exception(%.*s)\n", e.toString() );
	} catch ( AssertError ae ) {
		printf( "Caught AssertError(%.*s)\n", ae.toString() );
	} catch ( Object o ) {
		printf( "Caught Object(%.*s)\n", o.toString() );
	}
    return EXIT_SUCCESS;
}



